
/*
 *<SOURCE_HEADER>
 *
 *  <NAME>
 *    utils.c
 *  </NAME>
 *  <AUTHOR>
 *    Anthony R. Cassandra
 *  </AUTHOR>
 *  <CREATE_DATE>
 *    July, 1998
 *  </CREATE_DATE>
 *
 *  <RCS_KEYWORD>
 *    $RCSfile: utils.c,v $
 *    $Source: /u/cvs/proj/pomdp-solve/src/utils.c,v $
 *    $Revision: 1.10 $
 *    $Date: 2004/12/03 03:06:22 $
 *  </RCS_KEYWORD>
 *
 *  <COPYRIGHT>
 *
 *    2004, Anthony R. Cassandra
 *
 *    All Rights Reserved
 *                          
 *    Permission to use, copy, modify, and distribute this software and its
 *    documentation for any purpose other than its incorporation into a
 *    commercial product is hereby granted without fee, provided that the
 *    above copyright notice appear in all copies and that both that
 *    copyright notice and this permission notice appear in supporting
 *    documentation.
 * 
 *    ANTHONY CASSANDRA DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 *    INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR ANY
 *    PARTICULAR PURPOSE.  IN NO EVENT SHALL ANTHONY CASSANDRA BE LIABLE FOR
 *    ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 *    WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 *    ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 *    OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *  </COPYRIGHT>
 *
 *</SOURCE_HEADER>
 */

/*
 * Utility routines that require pomdp-solve code, but do not really
 * have anything directly to do with solving a POMDP. e.g., debugging,
 * solution undertsanding. 
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "mdp/mdp.h"

#include "global.h"
#include "params.h"
#include "pomdp.h"
#include "alpha.h"
#include "pg.h"
#include "belief.h"

double gUtilEpsilon = 1.0e-9;

/**********************************************************************/
/*
  This routine will read in  abelief list, and alpha list and then
  output another file containing the belief state, the best alpha
  vector for this state, the alpha vector number (order from alpha
  file, and the value that is attained.  It will also output the value
  for all of the other vectors.
*/
void
UTIL_mapBeliefList( char *belief_filename,
			char *alpha_filename,
			char *map_filename ) 
{
  BeliefList belief_list, b_ptr;
  AlphaList alpha_list, a_ptr, best_alpha;
  double *value, best_value;
  FILE *file;
  int a_num, i, best_count, *best_num;
  int *unique_best;
  int *best_for_belief_count;

   if ((file = fopen( map_filename , "w")) == NULL) {
     fprintf(gStdErrFile, 
             "** Error: The mapping file: %s cannot be opened.\n",
             map_filename);
     return;
   }

  belief_list = readBeliefList( belief_filename, 0 );
  if ( belief_list == NULL ) {
    fclose( file );
    return;
  }

  alpha_list = readAlphaList( alpha_filename, 0 );
  if ( alpha_list == NULL ) {
    destroyBeliefList( belief_list );
    fclose( file );
    return;
  }

  /* Print the file preamble */
  fprintf( file, "Belief Mapping File\n\n" );
  fprintf( file, "Belief File: %s (%d points)\n", 
		 belief_filename, sizeBeliefList( belief_list ));
  fprintf( file, "Alpha File: %s (%d vectors)\n\n", 
		 alpha_filename, sizeAlphaList( alpha_list ));
  
  value = (double *) XMALLOC( sizeAlphaList( alpha_list )
						* sizeof( double ));
  
  /* FIXME: Assume there will not be more than 5000 ties. */
  best_num =  (int *) XMALLOC( 5000 * sizeof( double ));

  /* We are also going to track which of the alpha vectors are
	uniquely the best for  agiven belief state.
  */
  unique_best =  (int *) XMALLOC( sizeAlphaList( alpha_list )
						    * sizeof( int ) );

  /* Also tracked is the number of belief states that each alpha
	vector is good for. */
  best_for_belief_count =  (int *) XMALLOC(  sizeAlphaList( alpha_list )
									* sizeof( int ) );

  /* initialization */
  for ( i = 0; i < sizeAlphaList( alpha_list ); i++) {
    unique_best[i] = 0;
    best_for_belief_count[i] = 0;
  }

  b_ptr = belief_list;
  while( b_ptr != NULL ) {
    
    best_value = worstPossibleValue();
    best_count = 0;
    best_alpha = NULL;
    a_num = 0;
    a_ptr = alpha_list->head;
    while( a_ptr != NULL ) {
	 
	 /* Get dot product value */
	 value[a_num] = 0.0;
	 for ( i = 0; i < gNumStates; i++)
	   value[a_num] += b_ptr->b[i] * a_ptr->alpha[i];

	 if ( GreaterThan( value[a_num], best_value, gUtilEpsilon )) {
	   best_value = value[a_num];
	   best_alpha = a_ptr;
	   best_num[0] = a_ptr->id;
	   best_count = 1;
	 }
	 else if ( Equal( value[a_num], best_value, gUtilEpsilon )) {
	   best_num[best_count] = a_ptr->id;
	   best_count++;
	 }

	 a_num++;
	 a_ptr = a_ptr->next;
    } /* while b_ptr != NULL */

    /* Now we are ready to output */

    fprintf( file, "Belief: " );
    for ( i = 0; i < gNumStates; i++ ) 
	 fprintf( file, "%.*lf ", BELIEF_FILE_DECIMAL_PRECISION,
			b_ptr->b[i] );
    fprintf( file, "\n" );

    fprintf( file, "Best Alpha Num(s): " );
    for ( i = 0; i < best_count; i++ ) {
	 fprintf( file, "%d ", best_num[i]) ;
	 best_for_belief_count[best_num[i]]++;
    }
    fprintf( file, "\n" );

    if ( best_count == 1 )
	 unique_best[best_num[0]]++;

    fprintf( file, "Best Alpha Value: %.*lf\n",
		   BELIEF_FILE_DECIMAL_PRECISION, best_value );

    fprintf( file, "Best Alpha: " );
    for ( i = 0; i < gNumStates; i++ )
	 fprintf( file, "%.*lf ", ALPHA_FILE_DECIMAL_PRECISION,
			best_alpha->alpha[i] );
    fprintf( file, "\n" );

    fprintf( file, "All Alpha Values: " );
    for ( i = 0; i < sizeAlphaList( alpha_list ); i++ ) 
	 fprintf( file, "%.*lf ", BELIEF_FILE_DECIMAL_PRECISION,
			value[i] );
    fprintf( file, "\n\n" );

    b_ptr = b_ptr->next;
  } /* while b_ptr != NULL */

  fprintf( file, "\nExtra Alpha Statistics:\n\n" );
  fprintf( file, "\n%12s%12s%12s\n", "AlphaID", "UniqueBest", "TotalBest" );

  for ( i = 0; i < sizeAlphaList( alpha_list ); i++) {
    fprintf( file, "%12d%12d%12d\n", 
		   i, unique_best[i], best_for_belief_count[i]  );
  }

  fprintf( file, "\nAll Alpha Vectors:\n\n" );
  displayAlphaList( file, alpha_list );

  XFREE( value );
  XFREE( best_num );
  XFREE( unique_best );
  XFREE( best_for_belief_count );

  destroyBeliefList( belief_list );
  destroyAlphaList( alpha_list );

}  /* UTIL_mapBeliefList */
/**********************************************************************/
/*
  This routine will read in an alpha list, sort it and then write it
  back out.  
*/
void
UTIL_sortAlphaFile( char *in_alpha_filename,
			char *out_alpha_filename ) 
{
  AlphaList in_alpha_list, out_alpha_list;

  in_alpha_list = readAlphaList( in_alpha_filename, 0 );
  if ( in_alpha_list == NULL ) {
    fprintf(gStdErrFile, 
		  "** Error: The alpha file: %s cannot be read.\n",
		  in_alpha_filename);
    return;
  }

  sortAlphaList( in_alpha_list );

  saveAlphaList( in_alpha_list, out_alpha_filename );

  destroyAlphaList( in_alpha_list );

}  /* UTIL_sortAlphaFile */

/**********************************************************************/
/*
  This routine will read in an alpha list, sort it and then write it
  back out.  
*/
void
UTIL_purgeAlphaFile( char *in_alpha_filename,
				 char *out_alpha_filename,
				 PomdpSolveParams param ) 
{
  AlphaList in_alpha_list;

  in_alpha_list = readAlphaList( in_alpha_filename, 0 );
  if ( in_alpha_list == NULL ) {
    fprintf(gStdErrFile, 
		  "** Error: The alpha file: %s cannot be read.\n",
		  in_alpha_filename);
    return;
  }

  fprintf( param->report_file, 
		 "Input Alpha Count: %d\n",
            sizeAlphaList(in_alpha_list)  );

  purgeAlphaList( in_alpha_list, 
                  param->q_purge_option, 
                  param );

  sortAlphaList( in_alpha_list );

  fprintf( param->report_file, 
		 "Ouput Alpha Count: %d\n",
            sizeAlphaList(in_alpha_list)  );

  saveAlphaList( in_alpha_list, out_alpha_filename );

  destroyAlphaList( in_alpha_list );

}  /* UTIL_purgeAlphaFile */

/**********************************************************************/
/***********   Match Alpha Routines     *******************************/
/**********************************************************************/

/**********************************************************************/
double
UTIL_measureAlphaDiff( double *src_alpha, 
				   double *dest_alpha )
{
  /*
    Compares two alpha vectors and returns a measure of their
    difference. Difference measure is always a positive number, where
    the smaller it is the better.

    Not clear at this time what the best
    comparisons is, but generally, we want to associate the two vectors
    from the two lists that have "the closest slope and values".  I
    believe that wanting both the closest slope and the closest values
    leads to some difficult to define metric.

    Some things to think about:

      Using the maximal component-wise distance (and taking the vector
      that produces the smallest max) [ This is the one that is
      currently implemented below. ]

      Using the total component-wise difference (which is equivalent to
      the average copmponent-wise difference since all vectors have the
      same number of components.)

    I think the best thing is to compute the volume of space between two
    vectors over the area of belief space where one or the other
    dominates.  However, this will require a fair amount of linear
    algebra as we will need to find out the intersection points, and
    then set up a multi-variate integral over some bounded region.
    Because everything is linear, the solution form may have some nice
    properties, but at trhe moment, I have not delved deep enough into
    the math to work this out.
   */
  int s;
  double cur_diff;
  double max_diff = -1.0*HUGE_VAL; /* max component-wise difference */

  for ( s = 0; s < gNumStates; s++ ) {

    cur_diff = fabs( src_alpha[s] - dest_alpha[s] );
    
    if ( cur_diff > max_diff )
	 max_diff = cur_diff;

  }  /* for s */
  
  return max_diff;
  
} /* UTIL_measureAlphaDiff */

/**********************************************************************/
int
UTIL_matchAlpha( double *src_alpha, 
		  AlphaList dest_alpha_list )
{
  /*
    Will take a source (src) alpha vector and will find the best
    matching vector in the destination (dest) list.  It will return
    the index of the vector in dest list that best matches.
    It will also fill in statistically information about the match in
    the MatchAlphaStats structure.
  */
  AlphaList walk_ptr;
  int cur_idx, best_idx;
  double cur_diff;
  double best_diff = HUGE_VAL;

  cur_idx = 0;
  walk_ptr = dest_alpha_list->head;
  while( walk_ptr != NULL ) {

    cur_diff = UTIL_measureAlphaDiff( src_alpha, walk_ptr->alpha );
    
    if ( cur_diff < best_diff )
	 {
	   best_idx = cur_idx;
	   best_diff = cur_diff;
	 }

    walk_ptr = walk_ptr->next;
    cur_idx++;
  }  /* while walk_ptr */

  return best_idx;

} /* matchAlpha */

/**********************************************************************/
int* 
UTIL_matchAlphaLists( AlphaList src_alpha_list, 
				  AlphaList dest_alpha_list )
{
  /*
    Will take a source (src) alpha list and create a mapping (of
    indices) from this list to the best matching vector in the
    destination (dest) list.  It will return a vector whose index is
    that of the index of the src alpha list, and whose value is the
    index of best matching alpha in the dest list.  It will also fill
    in statistically information about the match in the
    MatchAlphaListStats structure.
   */
  AlphaList walk_ptr;
  int from_idx, to_idx;
  int *idx_map;

  /* We will map one vector at a time in alpha1, so these arrays to
	track differences only need to be as large as the alpha2 list.
  */
  idx_map = (int *) XMALLOC( sizeof( int ) * sizeAlphaList(src_alpha_list) );

  from_idx = 0;
  walk_ptr = src_alpha_list->head;
  while( walk_ptr != NULL ) {

    to_idx = UTIL_matchAlpha( walk_ptr->alpha, dest_alpha_list );
    
    idx_map[from_idx] = to_idx;

    walk_ptr = walk_ptr->next;
    from_idx++;
  }  /* while walk_ptr */

  return idx_map;

} /* UTIL_matchAlphaLists */


/**********************************************************************/
void
UTIL_relinkPolicyGraph( char *alpha_filename,
				    char *pg_filename,
				    char *prev_alpha_filename,
				    char *out_pg_filename ) 
{
  /*
    This routine will create a policy graph that is a relinking of the
    input policy graph.  This is used when you want to use a policy
    graph from a finite horizon solution, where you do not have the
    same set of vectors from one iteration to the next.

    Here, you need to have the alpha and pg file for the horizon 'h'
    solution, as well as need to provide the alpha file from the 'h-1'
    iteration. The reason is that the horizon 'h' policy graph has
    edges that points into the 'h-1' policy graph.  These pointers are
    indices of the policy graph nodes in 'h-1'. However, the policy
    graph nodes indices are identical to the indices corresponding to
    the order of the 'h-1' alpha list.

    therefore, we find the closest match from the 'h-1' vectors to the
    'h' vectors, and then reset the 'h' pg pointers to point to the
    best match in the 'h' alpha list rathjer than those in the 'h-1'
    alpha list.
   */
  AlphaList alpha_list, prev_alpha_list;
  int *alpha_map;
  PG pg;

  alpha_list = readAlphaList( alpha_filename, 0 );
  if ( alpha_list == NULL ) {
    fprintf(gStdErrFile, 
		  "** Error: The alpha1 file: %s cannot be read.\n",
		  alpha_filename);
    return;
  }

  prev_alpha_list = readAlphaList( prev_alpha_filename, 0 );
  if ( prev_alpha_list == NULL ) {
    fprintf(gStdErrFile, 
		  "** Error: The alpha2 file: %s cannot be read.\n",
		  prev_alpha_filename);
    return;
  }

  /* Don't run sanity checks after reading, since here we know we may
	not have a valid PG (the whole purpose of this relinking is to turn
	it into something that is valid).
  */
  pg = PG_read( pg_filename, FALSE );
  if ( pg == NULL ) {
    fprintf(gStdErrFile, 
		  "** Error: The pg1 file: %s cannot be read.\n",
		  pg_filename);
    return;
  }

  if ( sizeAlphaList( alpha_list ) != pg->num_nodes ) {
    fprintf(gStdErrFile, 
		  "** Error: The pg1 and alpha1 files are not compatible.\n",
		  pg_filename);
    return;
  }

  if ( gNumObservations != pg->num_obs ) {
    fprintf(gStdErrFile, 
		  "** Error: The pg1 file is not compatible with POMDP file.\n",
		  pg_filename);
    return;
  }

  /* Note the reversal of the list order in this call.  Though the
	current pg points into the previous pg, to relink, we want the
	mapping from the previous to the best match in the current, since
	we will be replacing the previous alpha with current ones.
  */
  alpha_map = UTIL_matchAlphaLists( prev_alpha_list, alpha_list );

  PG_relink( pg, alpha_map, sizeAlphaList( prev_alpha_list )-1 );

  PG_write( pg, out_pg_filename );

  /* clean up time */
  destroyAlphaList( alpha_list );
  PG_Destructor( pg );
  destroyAlphaList( prev_alpha_list );
  XFREE( alpha_map );


} /* UTIL_relinkPolicyGraph */

/**********************************************************************/
/***********   Compare Alpha Routines     *******************************/
/**********************************************************************/

/**********************************************************************/
/*
  Here we attempt to match vectors from one alpha list to vectors in
  another alpha list.  Not clear at this time what the best
  comparisons is, but generally, we want to associate the two vectors
  from the two lists that have "the closest slope and values".  I
  believe that wanting both the closest slope and the closest values
  leads to some difficult to define metric.

  Some things to think about:

    Using the maximal component-wise distance (and taking the vector
    that produces the smallest max) [ This is the one that is
    currently implemented below. ]

    Using the total component-wise difference (which is equivalent to
    the average copmponent-wise difference since all vectors have the
    same number of components.)

  I think the best thing is to compute the volume of space between two
  vectors over the area of belief space where one or the other
  dominates.  However, this will require a fair amount of linear
  algebra as we will need to find out the intersection points, and
  then set up a multi-variate integral over some bounded region.
  Because everything is linear, the solution form may have some nice
  properties, but at trhe moment, I have not delved deep enough into
  the math to work this out.

*/
void
UTIL_compareAlphaFiles( char *alpha1_filename,
				    char *alpha2_filename,
				    double epsilon,
				    char *out_filename ) 
{
  FILE *file;
  AlphaList alpha1_list, alpha2_list;
  AlphaList walk_ptr1, walk_ptr2;

  /* Stores the differences between vectors.  We will keep this as a
	sorted list and doa  simple insertion sort as we put things into
	it. */ 
  double *diff;  

  /* As we do the insertion sort of the 'diff' arrray, we also want to
	keep track of the indices of the vectors producing the
	differences.  We'll want to output the mapping from alpha vetcor
	to alpha vector along with the differences.
  */
  int *idx;

  int s, from_idx, to_idx, insert_idx, move_idx;
  double max_diff;

  alpha1_list = readAlphaList( alpha1_filename, 0 );
  if ( alpha1_list == NULL ) {
    fprintf(gStdErrFile, 
		  "** Error: The alpha1 file: %s cannot be read.\n",
		  alpha1_filename);
    return;
  }

  alpha2_list = readAlphaList( alpha2_filename, 0 );
  if ( alpha2_list == NULL ) {
    fprintf(gStdErrFile, 
		  "** Error: The alpha2 file: %s cannot be read.\n",
		  alpha2_filename);
    return;
  }

  if ((file = fopen( out_filename , "w")) == NULL) {
    fprintf(gStdErrFile, 
		  "** Error: The output file: %s cannot be opened.\n",
		  out_filename);
    return;
  }

  /* We will map one vector at a time in alpha1, so these arrays to
	track differences only need to be as large as the alpha2 list.
  */
  diff = (double *) XMALLOC( sizeof( double ) * sizeAlphaList(alpha2_list) );
  idx = (int *) XMALLOC( sizeof( int ) * sizeAlphaList(alpha2_list) );
  
  fprintf( file, "Alpha List Comparison File\n\n" );
  fprintf( file, "Alpha1 File: %s\n", alpha1_filename );
  fprintf( file, "Alpha2 File: %s\n", alpha2_filename );
  fprintf( file, "Epsilon: %e\n\n", epsilon );

  from_idx = 0;
  walk_ptr1 = alpha1_list->head;
  while( walk_ptr1 != NULL ) {

    to_idx = 0;
    walk_ptr2 = alpha2_list->head;
    while( walk_ptr2 != NULL ) {


	 max_diff = UTIL_measureAlphaDiff( walk_ptr1->alpha,
								walk_ptr2->alpha );
            
	 /* Now insert this into the ordered list. (note that 'to_idx'
	    also happens to be the number of values already added to the
	    list. */

	 /* Find insertion_point */
	 for ( insert_idx = 0; insert_idx < to_idx; insert_idx++ )
	   if ( diff[insert_idx] > max_diff )
		break;

	 /* Move data down in list to accomodate new entry. (note we need
	    to move down both the diff and idx arrays). */
	 for ( move_idx = (to_idx-1); move_idx >= insert_idx; move_idx-- ) {
	   diff[move_idx+1] = diff[move_idx];
	   idx[move_idx+1] = idx[move_idx];
	 } /* for move_idx */

	 /* Now we can insert */
	 diff[insert_idx] = max_diff;
	 idx[insert_idx] = to_idx;

	 walk_ptr2 = walk_ptr2->next;
	 to_idx++;
    }  /* while walk_ptr2 */

    /* Now we output all those whose max_diff values is less than
	  epsilon. Note that in case there is *no* matching vector less
	  than epsilon, we output the lowest value, but flag it so it is
	  explict. */

    fprintf( file, "Index %d", from_idx );

    if ( diff[0] > epsilon ) {
	 fprintf( file, ": NONE, best is " );
    }

    fprintf( file, ": %d [%e] ", idx[0], diff[0] );
    
    for ( s = 1; s < sizeAlphaList(alpha2_list); s++ ) {

	 if ( diff[s] > epsilon ) 
	   break;

	 fprintf( file, ": %d [%e] ", idx[s], diff[s] );
	 
    } /* for s */

    fprintf( file, "\n" );
 


    walk_ptr1 = walk_ptr1->next;
    from_idx++;
  }  /* while walk_ptr1 */
  
  XFREE( diff );
  XFREE( idx );

  destroyAlphaList( alpha1_list );
  destroyAlphaList( alpha2_list );

}  /* UTIL_compareAlphaFiles */

/**********************************************************************/
/***********   Belief Update Routines   *******************************/
/**********************************************************************/

/**********************************************************************/
int 
UTIL_getLine( FILE *fh, char *string, int max)
{
   register int i;
   register int c;

   max--; 
   i = 0;

   while (( (c = getc(fh)) != EOF )
		&& ( c != '\n'))
	{
	  if ( i < max ) 
	    string[i++] = c;
	}

   string[i] = '\0';
   
   return (( c == EOF ) && (i == 0 ) ? EOF : i);
}

/**********************************************************************/
void 
UTIL_displayBelief( double *b )
{
  int i;

  fprintf( stdout, "[" );
  for ( i = 0; i < gNumStates; i++ ) 
    if ( i == 0 )
	 fprintf( stdout, " %.6lf", b[i] );
    else
	 fprintf( stdout, ", %.6lf", b[i] );
  fprintf( stdout, " ]\n" );

} /* UTIL_displayBelief */

/**********************************************************************/
void 
UTIL_doBeliefUpdates()
{
  /*
   * This function serves as a belief state updating module.  After
   * reading in a POMDP file specification it will continually
   * compute the next belief state given the current action and
   * observation.
   *
   * This program wil receive its input on stdin, and will consist of
   * action and observation.  The current belief state is initialized
   * from the POMDP file specification, or defaults to uniform if not
   * present. This module will track the belief state internally, thus
   * only requiring the action and observation to be input.
   *
   * The input action and observation must be given as an integre
   * oridinal value of the actions andobservations in the POMDP
   * file. This would be the index number corresponding to the
   * positional location of a mnemonic in the list (or simply the
   * 'i'th action/observation if no mnemomic list existed in the POMDP
   * file.)
   *
   * The output belief states will conform to the python default
   * syntax for array output: delimited by square brackets, with comma
   * separators between the elements in the array.
   *
   * All input and output will be line-based text.  Thus, the input
   * should be on a single line, e.g.
   *
   *    action  obs
   *
   * Inputting the line consisting only of the word 'exit' will cause the
   * program to terminate.
   *
   * The output will be the next belief state and will appear on a single
   * line as: 
   *
   *    [ O1, O2, ..., On ]
   *
   * The program will output the resulting belief state to stdout.  If
   * there are input errors, then a message will be displayed on
   * stderr and the return value will be an empty list ('[]'): An
   * input error will function as a no-op, so that the next input will
   * use the last good belief state in its update calculation.
   *
   * The one other possible output is a vector of all zeroes. This may
   * be the result of an input error, or more likely the result of a
   * modeling error (problem in POMDP file).  This vector of all
   * zeroes is generated when the program determines that based on the
   * current belief state and action taken, the observation given is
   * not possible (has zero probability).  This means that something
   * is inconsistent in the sense that something is generating an
   * observation that the model has decalred impossible.
   *
   */ 
   
  long action, obs;
  double *prev_belief, *next_belief;

  char line[1024];
  char *post_action_ptr;
  char *post_obs_ptr;

  char line_init[255];

  prev_belief = newBelief();
  next_belief = newBelief();

  copyBelief( prev_belief, gInitialBelief );

  while ( 1 ) 
    {
	 UTIL_displayBelief( prev_belief );

	 /* UTIL_getline() returns -1 on EOF (and other errors). */
	 if ( UTIL_getLine( stdin, line, 1023 ) == EOF )
	   break;

	 sscanf( line, "%s", line_init );

	 if ( strcmp( line_init, "exit" ) == 0 )
	   break;
	 
	 action = strtol( line, &post_action_ptr, 10 );

	 if ( post_action_ptr == line )   /* nothing parsed */
	   {
		fprintf( stderr, "Bad line input.\n" );
		fprintf( stdout, "[]\n" );
		continue;
	   }
	
	 if (( action < 0 ) || ( action >= gNumActions ))
	 {
		fprintf( stderr, "Bad action index.\n" );
		fprintf( stdout, "[]\n" );
		continue;
	 }

	 obs = strtol( post_action_ptr, &post_obs_ptr, 10 );

	 if ( post_obs_ptr == post_action_ptr )   /* nothing parsed */
	   {
		fprintf( stderr, "Bad line input.\n" );
		fprintf( stdout, "[]\n" );
		continue;
	   }
	
	 if (( obs < 0 ) || ( obs >= gNumObservations ))
	 {
		fprintf( stderr, "Bad observation index.\n" );
		fprintf( stdout, "[]\n" );
		continue;
	 }

	 /* Make sure there is not extraneous stuff around. */
	 while ( post_obs_ptr[0] != '\0' )
	   {
		if (( post_obs_ptr[0] != ' ' )
		    && ( post_obs_ptr[0] != '\t' )
		    && ( post_obs_ptr[0] != '\r' )
		    && ( post_obs_ptr[0] != '\n' ))
		  break;

		post_obs_ptr++;
	   }

	 if ( post_obs_ptr[0] != '\0' )
	   {
		fprintf( stderr, "Bad line input.\n" );
		fprintf( stdout, "[]\n" );
		continue;
	   }
 
	 transformBeliefState( prev_belief, next_belief, action, obs );

	 copyBelief( prev_belief, next_belief );
    }

} /* UTIL_doBeliefUpdates */

/**********************************************************************/
